/*
 * Copyright (c) 2015 by ELECTRONIC ASSEMBLY <technik@lcd-module.de>
 * EA EPA20-A (SSD1606) software library for arduino.
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 */

#include <Arduino.h>
#include <SPI.h>
#include <avr/pgmspace.h>

#include "epa20.h"

#define LUT_LEN 90
const unsigned  char lut_data [LUT_LEN]=
{
	0x82,0x00,0x00,0x00,0xAA,0x00,0x00,0x00,0xAA,
	0xAA,0x00,0x00,0xAA,0xAA,0xAA,0x00,0x55,0xAA,
	0xAA,0x00,0x55,0x55,0x55,0x55,0xAA,0xAA,0xAA,
	0xAA,0x55,0x55,0x55,0x55,0xAA,0xAA,0xAA,0xAA,
	0x15,0x15,0x15,0x15,0x05,0x05,0x05,0x05,0x01,
	0x01,0x01,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x41,
	0x45,0xF1,0xFF,0x5F,0x55,0x01,0x00,0x00,0x00
};

//----------------------------------------------------public Functions----------------------------------------------------
//Please use these functions in your sketch

/*----------------------------
Func: EA EPA20-A init
Desc: Initializes SPI Hardware/Software and EA EPA20-A
Vars: CS-Pin, MOSI-Pin, SCK-Pin (MOSI=MISO=SCK Hardware else Software), A0-Pin (high=data, low=command), p_res = Reset-Pin
------------------------------*/
void epa20::initialize(byte p_cs, byte p_si, byte p_so, byte p_clk, byte p_a0, byte p_res) 
{
	epa20::p_a0 = p_a0;
	pinMode(p_a0, OUTPUT);
	spi_initialize(p_cs, p_si, p_so, p_clk); //init SPI to Mode 3

	//perform a Reset
	pinMode(p_res, OUTPUT);
	digitalWrite(p_res, LOW);
	delay(1);
	digitalWrite(p_res, HIGH);
	delay(1);
	
	init_displ();
}

/*----------------------------
Func: update
Desc: Updates the screen content / refresh
Vars: ---
------------------------------*/
void epa20::update(void) 
{
	init_displ();  //disable deep sleep mode
	command(0x20); //booster on and update
	while(digitalRead(p_busy))
	 ;
	
	command(0x22); //booster disable
	data(0x03);
	command(0x20);
	command(0x10); //enter deep sleep mode
	data(0x01);
}

/*----------------------------
Func: clear_display
Desc: clears the entire Epaper-Display
Vars: ---
------------------------------*/
void epa20::clear(void) 
{
	unsigned int i;
	set_window(0,0, XPIXEL-1,YPIXEL-1); //set Window to whole screen
	command(0x24); //write data to ram
	for(i=0; i < XPIXEL * YPIXEL * 2 / 8; i++) //2 pixel per dot
		data(WHITE);  //Fill white
}

/*----------------------------
Func: string
Desc: shows string with selected font on position
Vars: x (0..171), y(0..71),  font adress in programm memory, stringarray, foreground color background color
------------------------------*/
void epa20::string(byte x, byte y, const byte *font_adress, const char *str, byte fg_color, byte bg_color)
{
	unsigned int dat, pos_array; 										//Postion of character data in memory array
	byte start_code, last_code, width_char, height_char, height_byte, bytes_p_char, page, i, x_char;
	
	y &= 0xFC;  //set y to page adresses (4 pixel boundary)	

	start_code 	 = pgm_read_byte(&font_adress[2]);  //get first defined character
	last_code	 = pgm_read_byte(&font_adress[3]);  //get last defined character
	width_char	 = pgm_read_byte(&font_adress[4]);  //widht in dots per char
	height_char	 = pgm_read_byte(&font_adress[5]);  //height in dots of one char
	height_byte   = pgm_read_byte(&font_adress[6]); //byte hight count per char
	bytes_p_char = pgm_read_byte(&font_adress[7]);  //bytes per char
	
	while(*str != 0) //print character by character
	{
		if((byte)*str < start_code || (byte)*str > last_code) //make sure data is valid
			str++;
		else
		{
			//calculate positon of ascii character in font array
			//bytes for header + (ascii - startcode) * bytes per char)
			pos_array = 8 + (unsigned int)((byte)*str++ - start_code) * bytes_p_char;
			page = y+height_char-1;	//mirror picture (start upper line page)
			for(i=0; i<height_byte; i++)
			{
				set_window(x,page-4,x+width_char,page);
				page-=8;
				
				command(0x24); //write data to ram
				for(x_char=0; x_char < width_char; x_char++) //print first page of the char
				{
					dat = b2w(pgm_read_byte(&font_adress[pos_array+x_char]), fg_color, bg_color); //stretch one byte to 8 dots (-> 2bytes)
					data(dat>>8);
					data(dat);
				}
				
				pos_array += width_char;
				
			}
		}
		x+=width_char;
	}
	
}

/*----------------------------
Func: rectangle
Desc: shows a pattern filled rectangle on the display
Vars: left lower corner, right upper corner, filling color (0xFF WHITE, 0x00 BLACK, 0x55 GRAY1, 0xAA GRAY2)
------------------------------*/
void epa20::rectangle(byte xs, byte ys, byte xe, byte ye, byte outline, byte filling)  
{
	byte buf;
	if(xs+1>xe)
	{
		buf = xe;
		xe = xs+1;
	}
	
	if(ys>ye)
	{
		buf = ye;
		ye = ys;
	}
	

	if(xe > XPIXEL-1)
		xe=XPIXEL-1;
	if(ye > YPIXEL-1)
		ye=YPIXEL-1;
		
	box(xs,ys,xe,ye, filling); //filling
	box(xs,ys,xe,ys, outline | (filling&0x0F)); //low line
	box(xs,ye,xe,ye, outline | (filling&0xF0)); //up line
	box(xs,ys,xs+1,ye, outline); //left line
	box(xe-1,ys,xe,ye, outline); //right line
	
	
}

/*----------------------------
Func: picture
Desc: shows a BLV-picture on the display (see BitMapEdit EA LCD-Tools (http://www.lcd-module.de/support.html))
Vars: x (0..171) and y(0..71), program memory adress of data, foreground color background color
------------------------------*/
void epa20::picture(byte x, byte y, const byte *pic_adress, byte fg_color, byte bg_color)  
{
	unsigned int i,n, width, height, dat, byte_cnt = 2, page;
		
	width = pgm_read_byte(&pic_adress[0]);
	height = pgm_read_byte(&pic_adress[1]);
	y &= 0xFC; //set y to page adresses (4 pixel boundary)	
	page = y+height-1;	//mirror picture (start upper line page)
	for(i=0; i<height/8; i++)
	{
		set_window(x,page-4,x+width-1,page);
		page -= 8;
		
		command(0x24); //write data to ram
		for(n=0; n < width; n++)
		{
			dat = b2w(pgm_read_byte(&pic_adress[byte_cnt++]), fg_color, bg_color); //stretch one byte to 8 dots (-> 2bytes)
			data(dat>>8);
			data(dat);
		}
		
	}
}

//----------------------------------------------------private Functions----------------------------------------------------
//normally you don't need those functions in your sketch
/*----------------------------
Func: init_displ
Desc: inits the display EA EPA20-A itself
Vars: ---
------------------------------*/
void epa20::init_displ (void)
{
	byte i;
	command (0x10);	//do not enter deep sleep mode
	data(0x00);
	command (0x11);	//data entry mode setting,0x03,Y increment,X increment
	data(0x03);
	command (0xF0);	//booster feedback selection,0x1F->internal feedback is used
	data(0x1F);		
	command (0x21);	//bypass the RAM data into the display,enable pass
	data(0x03);
	command (0x2C);	//write VCOM register
	data(0xA0);
	command (0x3C);	//board waveform, board voltage
	data(0x63);
	command (0x22);	//enable sequence, CLK->CP->
	data(0xC4);
	command (0x32);	//write LUT register
	for (i=0; i<LUT_LEN; i++)
		data(lut_data [i]);
}

/*----------------------------
Func: position
Desc: Sets RAM window
Vars: Startcoordinate, Endcoordinate
------------------------------*/
void epa20::set_window (byte xs, byte ys, byte xe, byte ye)
{
	ys = ys/4;
	ye = ye/4;	
	
	command(0x44); //set x adress (display is mirrored --> change y and x)
	data(ys&0x1F);
	data(ye&0x1F);
	command(0x4E);
	data(ys);
	
		
	command(0x45); //set y adress (display is mirrored --> change x and y)
	data(xs);
	data(xe);
	command(0x4F);
	data(xs);

}

/*----------------------------
Func: box
Desc: fills a rectangle
Vars: Startcoordinate, Endcoordinate, fillcolor
------------------------------*/
void epa20::box (byte xs, byte ys, byte xe, byte ye, byte filling)
{
	unsigned int i, dots;
	dots = ((xe-xs+1)*(ye-ys+4)*2) / 8;  //2 pixel per dot
	set_window(xs,ys, xe,ye);
	command(0x24); //write data to ram
	for(i=0; i < dots; i++) 
		data(filling);  //Fill white
}

/*----------------------------
Func: b2w
Desc: Byte strech to word
Vars: byte to strech, foregroundcolor, backgroundcolor, returns WORD
------------------------------*/
unsigned int epa20::b2w(byte b, byte fg, byte bg)
{
	byte i;
	unsigned int w = 0;
	fg &= 3; bg &= 3;	

	for(i = 0; i < 8; i++)
	{ 
		w <<= 2;
		w |= (b & 0x80) ? fg : bg;
		b <<= 1;
	}
	
	return w;
}


/*----------------------------
Func: command
Desc: Sends a command to the EPA-Display
Vars: data
------------------------------*/
void epa20::command(byte dat) 
{
	digitalWrite(p_a0, LOW);
	spi_put_byte(dat);
}

/*----------------------------
Func: data
Desc: Sends data to the EPA-Display
Vars: data
------------------------------*/
void epa20::data(byte dat) 
{
	 digitalWrite(p_a0, HIGH);
	 spi_put_byte(dat);
}

/*----------------------------
Func: spi_initialize
Desc: Initializes SPI Hardware/Software
Vars: CS-Pin, MOSI-Pin, SCK-Pin (MOSI=SCK Hardware else Software)
------------------------------*/
void epa20::spi_initialize(byte cs, byte si, byte so, byte clk) 
{
	//Set pin Configuration
	p_cs = cs;
	
	if(si == clk)
	{
		hardware = true;
		p_si = MOSI;
		p_clk = SCK;
		p_busy = MISO;
	}
	else 
	{
		hardware = false;
		p_si = si;
		p_clk = clk;
		p_busy = so;
	}
	
	// Set CS to deselct slaves
	digitalWrite(p_cs, HIGH);
	pinMode(p_cs, OUTPUT);
	
	// Set Data pin as output
	pinMode(p_si, OUTPUT);
	
	// Set Busy pin as input
	pinMode(p_busy, INPUT);
	
	// Set SPI-Mode 0: CLK idle low, rising edge, MSB first
	digitalWrite(p_clk, LOW);
	pinMode(p_clk, OUTPUT);
	if(hardware)
	{	
		SPI.begin();
		SPI.setBitOrder(MSBFIRST);
		SPI.setDataMode(SPI_MODE0);
		SPI.setClockDivider(SPI_CLOCK_DIV8);
	}
}

/*----------------------------
Func: spi_put_byte
Desc: Sends one Byte using CS
Vars: data
------------------------------*/
void epa20::spi_put_byte(byte dat) 
{
	digitalWrite(p_cs, LOW);
	spi_out(dat);
	digitalWrite(p_cs, HIGH);
}

/*----------------------------
Func: spi_put
Desc: Sends bytes using CS
Vars: ptr to data and len
------------------------------*/
void epa20::spi_put(byte *dat, int len) 
{
	digitalWrite(p_cs, LOW);
	do
	{
		spi_out(*dat++);
	}while(--len);
	
	digitalWrite(p_cs, HIGH);
}

/*----------------------------
Func: spi_out
Desc: Sends one Byte, no CS
Vars: data
------------------------------*/
void epa20::spi_out(byte dat) 
{
	byte i = 8;
	if(hardware) 
	{
		SPI.transfer(dat);
	}
	else 
	{
		do 
		{
			if(dat & 0x80)
				digitalWrite(p_si, HIGH);
			else
				digitalWrite(p_si, LOW);
			digitalWrite(p_clk, LOW);
			dat <<= 1;
			digitalWrite(p_clk, HIGH);
	  }while(--i);
	}
}

